'
' RETROGRA.BI
'
' "RETROGRA" is a graphics library for "retro" graphics, precisely spoken:
' limiting to 640x480 2D graphics with 256 colours in fullscreen and windowed
' mode, which is in fact more than the original VGA standard resp. beyond the
' capabilities of homecomputers
'
' replaces SCREENSIZE, SCREENLOCK and so on, "double-buffering" is forced
'
' It features further a graphical coloured console, video terminal (VT) style
' but A.O.T. with UTF-8 support
' brings a predefined palette based upon the console standard
'
' ment for retro-style games, also text-console-based, and other "simple"
' applications, a kind of "DOSBox" without the limitations of DOS
'
' This low resolution makes it much easier to create sprites and other graphics
'
' scales the output to fit better on high resoluted monitors
'
' Works with most ordinary FreeBASIC code, but do never use SCREENLOCK etc.
' instead of the rg versions
'
' ******************************************************************************
'
' rgSCREENNEW should create a new double-buffered canvas - 640x480 8 bpp
'   use 'rgSCREENNEW WINDOWED' to not start in fullscreen mode
'
' rgSCREENDELETE "destroys" the canvas somehow, like when exiting the program
'   => unnecessary
'
' rgSCREENSHOW - call this to make all of your graphic operations visible,
'   draws mouse pointer above all
'   waits for VSynch; use it instead of SCREENLOCK, SCREENUNLOCK and SCREENCOPY
'
' rgSHIFTPALETTE (int r, int g, int b)
'   shifts the colours of the default palette by RGB values of +/- 100(%)
'   0% restores the default palette values
'   only affects colour indicies 0-215, 216-239 may be changed "manually"
'   rbg+100 makes all white rbg-100 all black
'
' rgSETSCALING (int scaling)
'   sets video scaling factor, 0 disables software-scaling, 1 determines
'   best scaling factor (auto) - only even zoom-scaling supported
'
' rgLOADIMAGE (string filename, FB.IMAGE PTR dest_buffer[, int width, int height])
'   loads an image file using FBimage.lib (PNG, JPG...), converting it into RETROGRA
'   format for use with PUT
'   creates dest_buffer if 0, requires IMAGEDESTROY call then
'   width and height are limiting the size, smart attempts to fit a different sized
'   image, at least cutting to size
'   supports data folders/datafolder library, returns 0 on error
'
' rgLOADVTG (string filename, string dest_buffer[, int limitcolumns, int limitrows])
'   loads the first frame of a Video Text Graphic file as "image".
'   These are graphics made out of console characters, using control codes
'   and ESC-codes of the VT 2000 standard.
'
' rgSETMOUSE (int look, int x, int y)
'   can move the mouse cursor, selects a look, 0 for hide
'
' rgINITMOUSE ([string filename])
'   allows to use user-defined mouse-pointer, file must be 512x16 8bpp BMP file
'
' rgPRINT (string expression)
'   text console output - regard the control codes, especially ESC
'
' rgINPUT (string prompt, string var[, int limit=72[, int inputarea]])
'   primitive console input, only one-lined, always blocked at the page's end
'   overwrites background like rgPRINT, doesn't restore it
'   modal, i.e. blocks any other input until you hit Enter or Esc
'   only for strings, no validation check => write your own input function
'   prompt is an input prompt, input stored in var, limit is size limitation
'   in characters (UTF-8 !), inputarea limits the inputarea in characters,
'   you may therefore get an e.g. 16 character broad input box
'
' rgINKEY() is like INKEY$() but returns VT2000 compatible key codes
'   UTF-8 strings of special characters, VT2000 control codes like 13 for
'   Return & Enter, &hF0 to &hFF for function keys F1 to F16. Not distinguishing
'   doubly existing keys, e.g. number pad in number or cursor mode
'
' rgSHIFTKEYS([rgSHIFTKEY mask]) returns the shiftkey state of the last
'   rgINKEY() call
'   you may sum SHIFT, ALT, CMD, ALTGR, FN, CAPSLOCK for mask
'   returns a sum of those values;
'   #note: using "CMD" instead of "CONTROL"(Ctrl) - on Macs
'   "Control" is mostly an additional Alternate key, while "Command" has the
'   function of "Control". Therefore there is no "CONTROL", on Macs it will
'   equal ALTGR - which is an own shiftkey value, not necessarily Alt+Control
'   etc. - but also not necessarily you must press AltGr to get its
'   flag, depending on the platform you might get ALTGR+ALT+CMD at once
'   FN (special function key) may not work
'
' rgCLS
'   clears the screen using the current background colour of rgPRINT, also
'   empties the rgPRINT text buffer and assigning the current colours, places
'   the cursor at the "home" position
'
' rgRGB(int red, int green, int blue) / rgRGB(int &hRRGGBB)
'   returns the nearest matching console colour index from 0-215 to the given
'   red-green-blue values reaching from 0 to 100(%) resp. the commonly used
'   hexadecimal 000000-FFFFFF RGB value - no support of "&hRGB" though !
'   will return 240 ("permanent black") instead of 0
'
' AT(int column, int row) - creates an ESC P sequence for placing the cursor
'   using in rgPRINT expressions (like LOCATE in original FreeBASIC/PRINT)
'
' PRINTCOLOR(uint8 fgcolour, uint8 bgcolour) - returns an ESC p sequence
'   to set the printing colours, equals ESC("p", fgcolour, bgcolour)
'
' CRLF - the string "\r", maybe easier to type for some
'
' ESC(string letter[, uint8 a[,uint8 b]]) - creates an escape sequence:
'   ESC + letter + CHR(a, b), only allowed combinations
'
' Annotation: there are undocumented subs - those are internal and shouldn't
'             be used, for example all "rgc..."-SUBs (=> RetroGra-Console), but
'             not only
'
' Future enhancements:
' MESSAGEBOX - RETROGRA-GUI message box
' FILESELECT - RETROGRA-GUI file selection dialog
' rgDRAWTEXT(int x, int y, int size, string text[, colour[, decorations]])
'              - graphical output of single-lined UTF-8 text,
'              multicoloured / 216 colours, "all" sizes,
'              monospaced or proportional, scaled with antialiasing,
'              selectable bitmap- or "true type" fonts
'              does not execute control codes
'              size is always height of the line in pixels, character width
'              depends on this size value
'              colour is only working for monochrome fonts
'              text is always UTF-8 although the font file may only
'              cover ISO 8859-1 e.g.
' rgSETFONT(string) - selects a font (file)
' rgTEXTLENGTH(int size, string text) - returns width of a text without
'              putting it out
'
' rgDRAW(x, y, colour[, count]) - draws a point or polygone line supporting
'              antialiasing and transparency(alpha)
' rgPOLYGONE(x(), y(), colour[, count]) - draws a filled polygone
' rgLINE(x1,y1,x2,y2,colour,...) - draws one line with "styles" (arrows, dotted)
' rgBOX(x1,y1,x2,y2,colour,...) - draws a box line (also line patterns)
' rgRBOX - box with rounded edges
' rgFBOX - filled box (
' rgRFBOX - filled rbox
' rgELLIPSE
' rgFELLIPSE
'
' ******************************************************************************
'
' GRAPHICAL COLOURED CONSOLE
'
' the console output is based on historic terminals, where text output was
' exchangeable between monitors and printers => you might print also text
' on a fictive colour printer this way
' It's further capable of creating coloured character-based graphics like
' used by ZX-81 and C64 or "teletext"/"ceefax". Although limited, it
' outperforms clearly "ASCII art", which only uses ASCII characters to form
' graphics and mostly black-white
'
' It's part of a pseudo-standard you may call "VT-2000", derived from
' "VT-52". The 8-bit colour palette could be translated also in a greyscale,
' 4 bit and monochrome, based on "nearest colour". Many text decorations and
' features might not be supported and ignored, normally returning an error
' code or failed flag. Supports UTF-8.
'
' The font is monospaced with a fixed count of characters per line and lines
' per screen; all characters have the same size except some double sized
' characters like CJK characters, occupying 2 cells per char. Scrolls
' vertically by losing the screen content.
' In Retrogra limited to 80 characters per 30 lines on the screen only,
' therefore only 8x16 pixel sized cells and no space for many text decorations
' and looking REALLY "retro"...
'
' Each character/field has only a foreground and a background colour, but
' individually out of 256 colours simultanously
'
' Control Codes:
' Like ASCII, Codes below 0x20/32 (Space) have special functions and don't
' put out characters. Most important 27 - ESC, "Escape" also '\e': this is
' a code for very complex so called "escape sequences", which might cause
' most trouble if appearing unintended in strings
' The Nullbyte marks the end of a string, therefore stops the string processing
' even if it doesn't demand a strings' length
' The control codes aren't following the ANSI ASCII standard further -
' most important difference is Code 8 instead of 9 for the 8-character-wide
' tabulator
' although 13/0x0D/'\n' works as return/line break, the outdated sequence 13+10
' does not => this equals 11+2 instead ("carriage return" + "line feed")
' so use '\n' for line breaks
' use code 0x1A/26 to access the control characters' graphic, followed by
' the about 32 higher character code
' all multi-character codes have to be together in the same string; you can't
' send "parameter bytes" or data later with the next PRINT call
'
' Note, that all characters above code 127 have to be UTF-8 encoded always,
' a Nullbyte always ends the string/output everywhere.

DECLARE SUB rgINITMOUSE (BYVAL AS STRING="")
DECLARE SUB rgINITCONSOLE
DECLARE SUB RGSCREENSHOW
DECLARE FUNCTION rgRGB OVERLOAD (AS INTEGER, AS INTEGER, AS INTEGER) AS UBYTE
DECLARE FUNCTION rgRGB OVERLOAD (AS INTEGER) AS UBYTE

#INCLUDE ONCE "fbgfx.bi"
#INCLUDE ONCE "fbext.bi"
#INCLUDE ONCE "utf8.bi"
#INCLUDE ONCE "datafolder.bi"

DIM SHARED AS LONG retrograpalette(256)
DIM SHARED AS INTEGER retrograscaling, retrogra_scrstartx, retrogra_scrstarty, retrogra_scalestartx, retrogra_scalestarty, retrogra_cwidth, retrogra_cheight
DIM SHARED AS DOUBLE retrogralastrefresh

DIM SHARED AS ANY PTR retrograscreenupdatethread
DIM SHARED AS boolean retrograstopthreads=FALSE

' rgSCREENNEW canvas flags
ENUM rgSCREENFLAG
FULLSCREEN=1,WINDOWED=0,FORCEDFULLSCREEN=3,FRAMELESS=4,UNSCALED=8
END ENUM

' mouse cursor stuff -->
' rgSETMOUSE form names
ENUM rgMOUSEFORM
DISABLE=0,ARROW=1,BRACKET=2,BUSY=3,FAINTCROSS=4,CROSS=5,CROSSHAIR=6,POINTING=7,DRAGGING=8,QUESTION=9,BLOCKED=10
END ENUM

DIM SHARED AS rgMOUSEFORM retrogramousepointer
DIM SHARED AS FB.IMAGE PTR retrogramousepointerbitmap(16)
DIM SHARED AS FB.IMAGE PTR retrogramousepointermask(16)
DIM SHARED AS FB.IMAGE PTR retrogramousebackbuffer ' buffers background of mouse pointer
DIM SHARED AS INTEGER retrogramousepointerxo(16),retrogramousepointeryo(16)



' graphical console stuff -->
DIM SHARED AS UBYTE PTR rgcfont ' buffer containing font data
DIM SHARED AS UInteger rgcfontpageoffset(512) ' offsets to Unicode "pages" 0x0000+, 0x0080+, 0x0100+, 0x0180+ ...
DIM SHARED AS UInteger rgcfontlinesize(512) ' offset/size of a line in the corr. "page"
DIM SHARED AS USHORT rgcflags ' console mode flags
'                     1 = vertical overflow - scrolling on (default)
'                     2 = line overflow on (default)
'                    32 = always visible (maybe blinking) cursor
'                    64 = reserved / no effect, see '192'
'                   128 = sound on (default) - only BEL(alert) when line overflow is blocked
'                   192 = full sound enviroment (typing sound on inputs, carriage return...)
'                         note: the respective control codes are putting out sounds independently
DIM SHARED AS INTEGER rgccurx, rgccury ' current console cursor position 0-79 / 0-29
DIM SHARED AS UBYTE rgcfgcolour, rgcbgcolour ' current console char foreground+background colour
DIM SHARED AS USHORT rgccharflags ' current console char flags; decorations a.o.t.
'                     1 = "not empty" / normal
'                     2 = bold
'                     4 = italic
'                     8 = underlined
'                    16 = cancelled
'                    32 = "disabled" / "grey"
'                    64 = outlined
'                   128 = inverted
'                   512 = superscript
'                   256 = subscript
'                  1024 = double width (occupying double of cells)
'                  2048 = double height
'                  4096 = compressed (only for proportional outputs)
'   16384, 32768, 49152 = "double char" (right, below, right+below)
TYPE rgcCELL
  flags AS USHORT
  fgcolour AS UBYTE
  bgcolour AS UBYTE
  char AS ULONG
END TYPE
DIM SHARED AS rgcCELL rgccell(80,30)

' the "function" CRLF... (line break)
#DEFINE CRLF CHR(13)

' #note: "normal" shift keys <256, "lock" keys >255
ENUM rgSHIFTKEY
NONE=0,SHIFT=1,ALT=2,CMD=4,ALTGR=8,FN=128,CAPSLOCK=256,NUMLOCK=512,SCROLLLOCK=1024
END ENUM
DIM SHARED AS rgSHIFTKEY rgcshiftkeys ' shift key state rgINPUT/rgINKEY


SUB rgSCREENUPDATETHREAD(BYVAL parameter AS ANY PTR)
  DO
    IF (TIMER-retrogralastrefresh)>0.125 THEN
      rgSCREENSHOW
    ENDIF
    SLEEP 25,1
  LOOP WHILE retrograstopthreads=FALSE
END SUB


SUB rgSCREENNEW (BYVAL flags AS rgSCREENFLAG=FULLSCREEN, BYVAL scaling AS INTEGER=1)
  ' flags: WINDOWED <-> FULLSCREEN,
  '        "FORCEDFULLSCREEN" creates a desktop sized canvas for fullscreen
  '        FRAMELESS = borderless window
  '        UNSCALED = perform no scaling => 640x480 window content
  ' scaling: forced scaling factor, 0 = no scaling, 1 = auto(default)
  '          if windowed mode, window size = scaling size, must not exceed monitor resolution
  DIM AS INTEGER c, r, g, b, dc

  ' *** creating colour console palette ***
  ' assigning RGB-scale colours 0-215
  ' in FreeBASIC format => 6 bits, 0-63 per colour channel only
  c=0
  FOR b=0 TO 63 STEP 12
    IF b=24 THEN b=27
    FOR g=0 TO 63 STEP 12
      IF g=24 THEN g=27
      FOR r=0 TO 63 STEP 12
        IF r=24 THEN r=27
        retrograpalette(c)=(b SHL 16) + (g SHL 8) + r
        c+=1
      NEXT
    NEXT
  NEXT
  ' free/user defined colours => dummy colour values
  FOR c=216 TO 238 STEP 2
    retrograpalette(c)=&h102030
    retrograpalette(c+1)=&h112233
  NEXT
  ' 0xF0-0xFF console special effect colours
  retrograpalette(240)=0 ' "permanent" black (turns to 0 in memory)
  ' also dummy colours
  FOR c=241 TO 254
    retrograpalette(c)=(((13*(c MOD 2))+50) SHL 16)+(13*(c MOD 2))+50
  NEXT
  retrograpalette(255)=&h303F3F ' inverting effect, dummy colour

  ' *** creating canvas ***
  DIM AS INTEGER res, w, h, d
  IF (flags AND FORCEDFULLSCREEN) THEN ' force desktop sized fullscreen 8 bpp with scaling
    res=0 ' don't try 640x480 in fullscreen
  ELSE
    res=SCREENLIST(8) ' list of 8 bpp fullscreen graphic modes
    WHILE res
      w=HIWORD(res)
      h=LOWORD(res)
      IF (w=640 AND h=480) THEN EXIT WHILE
      res=SCREENLIST
    WEND
  ENDIF

  IF res THEN ' fullscreen available for 640x480 8 bpp
    retrograscaling=0 ' scaling off
    IF (flags AND FULLSCREEN) THEN
      SCREENRES 640, 480, 8, 3, FB.GFX_FULLSCREEN + FB.GFX_NO_FRAME
    ELSE
      SCREENRES 640, 480, 8, 3, FB.GFX_WINDOWED
    ENDIF
    retrogra_cwidth=640
    retrogra_cheight=480
  ELSE ' 640x480 8 bpp not supported
    ' creating a desktop sized canvas for fullscreen
    '   or a window best scaled to desktop size
    SCREENINFO w, h, d ' get desktop resolution
    IF scaling THEN
      retrogra_scrstartx=0 ' start coords in virtual(640x480) screen
      retrogra_scrstarty=0
      retrogra_scalestartx=0
      retrogra_scalestarty=0
      retrograscaling=h/480 ' most likely, height doesn't fit as often in desktop resolutions like width do
      IF (w/640)<retrograscaling THEN retrograscaling=w/640 ' but otherwise...
      IF scaling<=retrograscaling THEN
        IF scaling>1 THEN ' not auto scaling => user demand
          retrograscaling=scaling
        ENDIF
      ELSE ' scaling exceeds screen size
        ' determine whether cutting loss is small
        IF (480-INT(h/scaling)<33 AND 640-INT(w/scaling)<33) THEN ' cut only up to 32 pixels
          retrograscaling=scaling
          retrogra_scrstartx=(640-INT(w/retrograscaling)) SHR 1 ' start copying original from these coords
          IF retrogra_scrstartx<1 THEN retrogra_scrstartx=0
          retrogra_scrstarty=(480-INT(h/retrograscaling)) SHR 1
          IF retrogra_scrstarty<1 THEN retrogra_scrstarty=0
        ENDIF ' ELSE keep the previously calculated value
      ENDIF

      ' determine offsets
      IF NOT (retrogra_scrstartx OR retrogra_scrstarty) THEN ' if this offsets>0 => starting at 0
        retrogra_scalestartx=(w-640*retrograscaling) SHR 1 ' centers output
        retrogra_scalestarty=(h-480*retrograscaling) SHR 1
      ENDIF
    ENDIF

    IF (flags AND FULLSCREEN) THEN ' fullscreen modes
      SCREENRES w, h, 8, 3, FB.GFX_FULLSCREEN + FB.GFX_NO_FRAME
      ' #todo: error 8 bpp failed / no canvas
    ELSE ' windowed modes
      w=640*retrograscaling
      h=480*retrograscaling
      IF (flags AND FRAMELESS) THEN ' window without decorations
        SCREENRES w, h, 8, 3, FB.GFX_NO_FRAME
      ELSE
        SCREENRES w, h, 8, 3
      ENDIF
      ' canvas has exact matching size
      retrogra_scalestartx=0
      retrogra_scalestarty=0
    ENDIF
    retrogra_cwidth=w
    retrogra_cheight=h
  ENDIF

  SCREENCOPY 2,0
  SCREENSET 2,0

  ' *** load mouse pointer graphics ***
  rgINITMOUSE

  ' *** init console / load console font ***
  rgcfont=0
  rgINITCONSOLE

  PALETTE USING retrograpalette

  ' start refresh-forcing thread
  retrograstopthreads=FALSE
  retrogralastrefresh=TIMER
  retrograscreenupdatethread=THREADCREATE(@rgSCREENUPDATETHREAD)
END SUB


SUB rgSCREENDELETE
  ' "deconstructor"

  IF retrograscreenupdatethread THEN
    'stop all threads
    retrograstopthreads=TRUE
    THREADWAIT(retrograscreenupdatethread)
  ENDIF

  IF SCREENPTR THEN ' rgSCREENNEW was successfully called
    ' free buffers
    IMAGEDESTROY retrogramousebackbuffer
    FOR i AS INTEGER=0 TO 15
      IMAGEDESTROY retrogramousepointerbitmap(i)
      IMAGEDESTROY retrogramousepointermask(i)
    NEXT
  ENDIF

  ' delete screens if reserved / close graphic window
  SCREEN 0
END SUB


SUB rgSCREENSHOW
  DIM AS INTEGER x, y, o, mx, my
  DIM AS BYTE c
  DIM AS BYTE PTR vga_screen, logical_screen

  IF (TIMER-retrogralastrefresh<0.005) THEN EXIT SUB ' avoid "ridiculous refresh rates"

  ' place mousecursor (if active)
  IF retrogramousepointer THEN
    SCREENSET 2,0
    GETMOUSE(mx, my)
    IF retrograscaling THEN
      mx/=retrograscaling
      my/=retrograscaling
      IF mx>639 THEN mx=639
      IF my>479 THEN my=479
      mx-=retrogramousepointerxo(retrogramousepointer-1)
      my-=retrogramousepointeryo(retrogramousepointer-1)
      IF (mx<0 OR my<0) THEN
        IF (mx<0 AND my<0) THEN
          GET (0,0)-(15,15),retrogramousebackbuffer
        ELSE
          IF mx<0 THEN
            GET (0,my)-(15,my+15),retrogramousebackbuffer
          ELSE
            GET (mx,0)-(mx+15,15),retrogramousebackbuffer
          ENDIF
        ENDIF
      ELSE
        GET (mx,my)-(mx+15,my+15),retrogramousebackbuffer
      ENDIF
      PUT (mx, my), retrogramousepointermask(retrogramousepointer-1),AND
      PUT (mx, my), retrogramousepointerbitmap(retrogramousepointer-1),OR
    ELSE
      IF mx>639 THEN mx=639
      IF my>479 THEN my=479
      GET (mx,my)-(mx+15,my+15),retrogramousebackbuffer
      PUT (mx, my), retrogramousepointermask(retrogramousepointer-1),AND
      PUT (mx, my), retrogramousepointerbitmap(retrogramousepointer-1),OR
    ENDIF
  ENDIF

  ' check whether scaling is done / perform scaling
  IF retrograscaling THEN
    ' getting video buffers, scaling-copy memory to memory
    vga_screen=SCREENPTR
    SCREENSET 1,0
    logical_screen=SCREENPTR

    ' #note: retrograscaling is either 0, 2 or higher/3-5 about
    '        if it's 1 or too high for the screen, it should really crash
    '        this must not happen
    IF retrograscaling=2 THEN
      ' scaling 2 - 2x2 copying
      FOR y=0 TO 479
        FOR x=0 TO 639
          c=*(vga_screen+y*retrogra_cwidth+x)
          o=((y SHL 1)+retrogra_scalestarty)*retrogra_cwidth+(x SHL 1)+retrogra_scalestartx
          *(logical_screen+o)=c
          o+=1
          *(logical_screen+o)=c
          o+=retrogra_cwidth
          o-=1
          *(logical_screen+o)=c
          o+=1
          *(logical_screen+o)=c
        NEXT
      NEXT
      SCREENSYNC
      SCREENCOPY 1,0
    ELSE
      ' scaling from 3 above - 3x3, 4x4 etc. copies
      DIM AS INTEGER x2, y2, screndx, screndy
      screndx=639-retrogra_scrstartx
      screndy=479-retrogra_scrstarty
      FOR y=retrogra_scrstarty TO screndy
        FOR x=retrogra_scrstartx TO screndx
          c=*(vga_screen+y*retrogra_cwidth+x)
          o=((y*retrograscaling)+retrogra_scalestarty)*retrogra_cwidth+(x*retrograscaling)+retrogra_scalestartx
          y2=retrograscaling
          DO
            x2=retrograscaling
            DO
              *(logical_screen+o)=c
              o+=1
              x2-=1
            LOOP WHILE x2
            o+=retrogra_cwidth
            o-=retrograscaling
            y2-=1
          LOOP WHILE y2
        NEXT
      NEXT
      SCREENSYNC
      SCREENCOPY 1,0 ' copy scaled to physical screen
    ENDIF
  ELSE
    SCREENSYNC
    SCREENCOPY 2,0 ' copy virtual to physical screen
  ENDIF
  SCREENSET 2,0 ' output in virtual screen
  retrogralastrefresh=TIMER

  ' remove mousecursor "sprite" (if active)
  IF retrogramousepointer THEN
    IF (mx<0) THEN mx=0
    IF (my<0) THEN my=0
    PUT (mx,my),retrogramousebackbuffer,PSET
  ENDIF
END SUB


SUB rgSETSCALING (BYVAL scaling AS INTEGER)
  IF scaling<1 THEN
    retrograscaling=0 ' scaling disabled
    ' #todo: adapt window size etc.
    EXIT SUB
  ENDIF
  IF scaling>1 THEN
    IF scaling=retrograscaling THEN EXIT SUB
  ENDIF

  IF scaling=1 THEN ' determine best scaling factor (auto)
    ' #todo: set retrograscaling, window size, offsets and so on
  ELSE
    ' #todo:
    ' check scaling to be valid, adapt scaling variables, resize window if windowed mode
    ' determine max screen size => 'scaling' multiples of 640 AND(!) 480 must fit in
    ' available screen
  ENDIF
END SUB


SUB rgSHIFTPALETTE (BYVAL rperc AS BYTE=0, BYVAL gperc AS BYTE=0, BYVAL bperc AS BYTE=0)
  DIM AS ULONG i, r, g, b
  DIM AS LONG newpal(256)

  FOR i=0 TO 215
    IF rperc=0 THEN ' keep value
      r=retrograpalette(i) AND &H3F
    ELSEIF rperc=100 THEN ' max. value (of white)
      r=&H3F
    ELSEIF rperc=-100 THEN ' min. value (of black)
      r=0
    ELSE
      r=retrograpalette(i) AND &H3F
      IF rperc>0 THEN
        r+=(63-r)*rperc/100
      ELSE
        r-=r*rperc/-100
      ENDIF
      r AND=63
    ENDIF
    IF gperc=0 THEN ' keep value
      g=retrograpalette(i) AND &H3F00
    ELSEIF gperc=100 THEN ' max. value (of white)
      g=&H3F00
    ELSEIF gperc=-100 THEN ' min. value (of black)
      g=0
    ELSE
      g=retrograpalette(i) AND &H3F00
      IF gperc>0 THEN
        g+=(&H3F00-g)*gperc/100
      ELSE
        g-=g*gperc/-100
      ENDIF
      g AND=&H3F00
    ENDIF
    IF bperc=0 THEN ' keep value
      b=retrograpalette(i) AND &H3F0000
    ELSEIF bperc=100 THEN ' max. value (of white)
      b=&H3F0000
    ELSEIF bperc=-100 THEN ' min. value (of black)
      b=0
    ELSE
      b=retrograpalette(i) AND &H3F0000
      IF bperc>0 THEN
        b+=(&H3F0000-b)*bperc/100
      ELSE
        b-=b*bperc/-100
      ENDIF
      b AND=&H3F0000
    ENDIF
    newpal(i)=r+g+b
  NEXT
  FOR i=216 TO 254
    newpal(i)=retrograpalette(i)
  NEXT
  PALETTE USING newpal
END SUB


FUNCTION rgLOADIMAGE (BYVAL filename AS STRING, BYREF dest_buffer AS FB.IMAGE PTR, BYVAL width_ AS INTEGER=0, BYVAL height AS INTEGER=0) AS FB.IMAGE PTR
  DIM AS ULONG PTR palette_
  DIM AS ULONG i, j, c
  DIM AS INTEGER x, y, lw
  DIM AS ULONG r, g, b

  palette_=ALLOCATE(256*4)
  IF palette_=0 THEN RETURN 0 ' really bad omen error

  IF DFLOADIMAGE(filename, dest_buffer, "", palette_) THEN
    ' convert to VT-2000 colours if not matching
    ' FB.IMAGE only has the pixelformat modes 8, 16 and 32 bpp, others are converted during loading
    IF dest_buffer->bpp=1 THEN
      ' 1-8 bpp (1 byte per pixel)
      FOR i=0 TO 215
        ' (palette format is 0x00RRGGBB)
        c=((i MOD 6)*&h330000)+(int((i MOD 36)/6)*&h3300)+(INT(i/36)*&h33)
        IF palette_[i]<>c THEN EXIT FOR
      NEXT
      IF i<>216 THEN
        ' palette not matching => effortful conversion of all pixels
        ' #todo: create conversion palette sourcecol(i) -> destcol(i)
        '        then exchange every pixel with assigned colour
        ' #note: use colour #240 instead of 0
      ENDIF
    ELSEIF dest_buffer->bpp=2 THEN ' 16 bpp
      lw=dest_buffer->width SHL 1
      FOR y=0 TO dest_buffer->height
        FOR x=0 TO dest_buffer->width
          c=PEEK(USHORT, dest_buffer+SIZEOF(fb.image)+y*32+x)
          r=c SHR 11
          g=(c SHR 5) AND &h3F
          b=c AND &1F
        NEXT
      NEXT
    ELSEIF dest_buffer->bpp=4 THEN ' 32 bpp
      DIM AS FB.IMAGE PTR buffer ' intermediate buffer
      DIM AS ULONG PTR p32
      DIM AS UBYTE PTR p8
      DIM AS INTEGER dummy, lw32, lw8
      x=dest_buffer->width
      y=dest_buffer->height
      buffer=IMAGECREATE(x, y, 0, 8)
      IMAGEINFO dest_buffer, dummy, dummy, dummy, lw32, p32
      IMAGEINFO buffer, dummy, dummy, dummy, lw8, p8
      lw32 SHR=2 ' to length in long words

      i=0
      j=0
      WHILE y
        y-=1
        lw=x
        WHILE lw
          lw-=1
          c=p32[i]
          r=int((((c AND &hFF0000) SHR 16)+25)/51)
          g=INT((((c AND &hFF00) SHR 8)+25)/51)
          b=INT(((c AND &hFF)+25)/51)
          p8[j]=b*36+g*6+r
          i+=1
          j+=1
        WEND
        IF (i MOD lw32) THEN
          i=INT(i/lw32)
          i+=1
          i*=lw32
        ENDIF
        IF (j MOD lw8) THEN
          j=INT(j/lw8)
          j+=1
          j*=lw8
        ENDIF
      WEND

      IMAGEDESTROY dest_buffer
      dest_buffer=buffer
    ENDIF
    DEALLOCATE palette_
    RETURN dest_buffer
  ENDIF

  RETURN 0 ' error
END FUNCTION


FUNCTION rgLOADVTG (BYVAL filename AS STRING, BYREF buffer AS STRING, BYVAL limitc AS INTEGER=0, BYVAL limitr AS INTEGER=0) AS STRING
  ' loads a VideoTerminalGraphic file
  DIM AS INTEGER i, j, h
  DIM AS STRING file="", line_

  ' #todo: / #note: has not been tested/debugged
  buffer=""
  IF DFLOAD(filename, file) THEN
    IF LEFT(file, 4)<>"VTG"+CHR(13) THEN ' header invalid
      RETURN ""
    ENDIF
    ' convert common newline codes to VT2000
    file=STRREPLACE(file, CHR(10,13), CHR(13))
    file=STRREPLACE(file, CHR(10), CHR(13))
    i=INSTR(file, CHR(13)+"height=")
    IF i THEN
      h=VAL(MID(file, i+8))
      IF h>limitr THEN h=limitr
    ELSE
      h=limitr
    ENDIF
    i=INSTR(file, CHR(13)+"FRAME")
    IF i THEN
      i+=1
      i=INSTR(i, file, CHR(13))
      IF i<1 THEN RETURN "" ' error / EOF
      i+=1
      j=INSTR(i, file, CHR(13))
      IF j<1 THEN RETURN "" ' error / EOF
      WHILE h
        line_=MID(file, i, j-i)
        IF line_="END\r" THEN EXIT WHILE ' end of frame
        buffer+=line_

        i=j
        j=INSTR(i, file, CHR(13))
        IF j<1 THEN EXIT WHILE ' end of file
        j+=1

        h-=1
      WEND
    ELSE
      ' no (valid) frame, invalid, error
      RETURN ""
    ENDIF
    ' #todo: limit width of each line (simply reaching from \r to \r)
    '        requires calculation of control codes, removing ESC codes (2-4 bytes long, depending...)
    '        and string size in UTF-8 (C2-DF 2 bytes per char, E0-EF 3 bytes, F0-F4 4 bytes,
    '        +double-width chars)
  ENDIF

  RETURN buffer
END FUNCTION


SUB rgINITMOUSE (BYVAL filename AS STRING="")
  IF retrogramousepointerbitmap(0) THEN EXIT SUB ' already initialised
  DIM AS INTEGER i, x, y, c
  DIM mousepointers AS FB.IMAGE PTR
  ' init buffer for pointer graphics - always 8 pointers 16x16
  mousepointers=IMAGECREATE(512,16,0,1)

  ' load pointer graphics BMP
  IF filename="" THEN ' no filename => load default
    rgLOADIMAGE("%FBDATA%retrogra/mouse.bmp", mousepointers)
  ELSE
    IF rgLOADIMAGE(filename, mousepointers)=FALSE THEN
      ' likely given file not found/failed => load default
      rgLOADIMAGE("%FBDATA%retrogra/mouse.bmp", mousepointers)
    ENDIF
  ENDIF

  retrogramousebackbuffer=IMAGECREATE(16,16,0,1) ' buffers background of mouse pointer
  ' create sprites from image, get hotspots ...
  FOR i=0 TO 15
    retrogramousepointerbitmap(i)=IMAGECREATE(16,16,0,1)
    retrogramousepointermask(i)=IMAGECREATE(16,16,0,1)
    GET mousepointers,(i*32,0)-(i*32+15,15),retrogramousepointerbitmap(i)
    GET mousepointers,(i*32+16,0)-(i*32+31,15),retrogramousepointermask(i)
    retrogramousepointerxo(i)=1
    retrogramousepointeryo(i)=1
    FOR y=0 TO 15
      FOR x=0 TO 15
        c=POINT(x,y,retrogramousepointermask(i))
        IF c=5 THEN ' hotspot opaque
          retrogramousepointerxo(i)=x
          retrogramousepointeryo(i)=y
          PSET retrogramousepointermask(i), (x, y), 0
          EXIT FOR, FOR
        ENDIF
        IF c=30 THEN ' hotspot transparent
          retrogramousepointerxo(i)=x
          retrogramousepointeryo(i)=y
          PSET retrogramousepointermask(i), (x, y), 255
          EXIT FOR, FOR
        ENDIF
      NEXT
    NEXT
  NEXT
  IMAGEDESTROY mousepointers
  SETMOUSE , , 0
  retrogramousepointer=1
END SUB


SUB rgINITCONSOLE
  DIM AS INTEGER i, x, y
  DIM AS UBYTE PTR address

  ' load console font file data into memory
  IF DFLOAD("%FBDATA%font/c16x8.fnt", rgcfont) THEN
    ' get Unicode "page" offsets
    address=rgcfont+&H80
    FOR i=0 TO 511
      rgcfontpageoffset(i)=PEEK(ULONG, address)
      address+=4
    NEXT
    ' get font data line width of each "page" / set offset to first char data
    FOR i=0 TO 511
      IF rgcfontpageoffset(i) THEN
        '  address=rgcfont+rgcfontpageoffset(i)
        rgcfontlinesize(i)=128 ' #todo: read data really, but FNT file not yet ready !
        rgcfontpageoffset(i)+=4
      ENDIF
    NEXT
  ELSE
    ' #todo: file open error => fatal
    rgcfontpageoffset(0)=0 ' definitely wrong value
  ENDIF

  ' set default/starting values
  rgccurx=0
  rgccury=0
  rgcfgcolour=0
  rgcbgcolour=215
  rgcflags=131 ' default: overflow+scrolling+BEL sound
  rgccharflags=1 ' default: normal / draw character

  ' "clear" "screen"
  ' default is black on white
  ' if it should start e.g. black, you have to overdraw the screen
  ' before the first rgSCREENSHOW call
  FOR y=0 TO 29
    FOR x=0 TO 79
      rgccell(x,y).flags=0
      rgccell(x,y).fgcolour=0 ' black
      rgccell(x,y).bgcolour=215 ' white
      rgccell(x,y).char=0
    NEXT
  NEXT
  LINE (0,0)-(639,479),215,BF
END SUB


SUB rgGETMOUSE (BYREF mx AS INTEGER, BYREF my AS INTEGER, BYREF wheel AS INTEGER=0, BYREF button AS INTEGER=0)
  GETMOUSE(mx, my, wheel, button)
  IF retrograscaling THEN
    mx/=retrograscaling
    my/=retrograscaling
  ENDIF
  IF mx>639 THEN mx=639
  IF my>479 THEN my=479
END SUB


SUB rgSETMOUSE (BYVAL form AS rgMOUSEFORM=-1, BYVAL x AS INTEGER=-1, BYVAL y AS INTEGER=-1)
  ' form: -1 illegal / don't change - 0-16 allowed
  '        0 invisible / disable
  '       standards:
  '        1 arrow / pointer
  '        2 doublebracket
  '        3 busy
  '        4 thin cross
  '        5 thick cross
  '        6 "crosshair"
  '        7 pointing hand
  '        8 dragging hand
  '        9 question mark
  '       10 blocked
  IF (form>=0 AND form<=16) THEN retrogramousepointer=form
  IF (x<>-1 OR y<>-1) THEN
    IF retrograscaling THEN
      IF x<>-1 THEN x*=retrograscaling
      IF y<>-1 THEN y*=retrograscaling
    ENDIF
    SETMOUSE x, y, 0
  ENDIF
END SUB


FUNCTION AT(BYVAL column AS byte, BYVAL row AS byte) AS STRING
  IF column>0 THEN
    IF column>80 THEN column=80 ' outside console area
  ELSE
    column=1 ' 0/negative are illegal values - no Nullbytes allowed also
  ENDIF
  IF row>0 THEN
    IF row>30 THEN row=30 ' outside console area
  ELSE
    row=1 ' 0/negative are illegal values - no Nullbytes allowed also
  ENDIF
  AT=CHR(27)+"P"+CHR(column,row)
END FUNCTION


FUNCTION PRINTCOLOR(BYVAL fgcolour AS UBYTE=250, BYVAL bgcolour AS UBYTE=250) AS STRING
  DIM AS STRING f, b
  DIM AS INTEGER i

  ' colour 250/0xFA: don't change colour
  IF fgcolour=250 THEN fgcolour=rgcfgcolour
  IF bgcolour=250 THEN bgcolour=rgcbgcolour

  IF fgcolour THEN
    IF fgcolour>127 THEN ' requires UTF-8 encoding
      f=CHR(((fgcolour SHR 6) OR 192), ((fgcolour AND 63)+128)) ' only 8 bit; 0-255 possible
    ELSE
      f=CHR(fgcolour)
    ENDIF
  ELSE ' colour 0 => translate to "permanent black" 0xF0/240 (also in UTF-8)
    f=CHR(195, 144)
  ENDIF
  IF bgcolour THEN
    IF bgcolour>127 THEN ' requires UTF-8 encoding
      b=CHR(((bgcolour SHR 6) OR 192), ((bgcolour AND 63)+128)) ' only 8 bit; 0-255 possible
    ELSE
      b=CHR(bgcolour)
    ENDIF
  ELSE ' colour 0 => translate to "permanent black" 0xF0/240 (also in UTF-8)
    b=CHR(195, 144)
  ENDIF
  PRINTCOLOR=CHR(27)+"p"+f+b
END FUNCTION


FUNCTION ESC(BYVAL letter AS STRING, BYVAL a AS INTEGER=0, BYVAL b AS INTEGER=0) AS STRING
  DIM AS STRING result=""
  DIM AS UBYTE c

  letter=LEFT(letter, 1)
  c=letter[0]
  IF a THEN IF a<0 OR a>255 THEN a=0 ' only UBYTE range allowed
  IF b THEN IF b<0 OR b>255 THEN b=0 ' only UBYTE range allowed
  IF c<65 OR (c>90 AND c<97) OR c>122 THEN RETURN "" ' error, must be ASCII letter

  ' only valid VT 2000 escape sequences
  SELECT CASE letter
    CASE "M" ' byte encoding: 0 ASCII, 1-16 ISO 8859, 127 UTF-8(16/32 autodetect)
      IF a<>127 and a>16 THEN RETURN "" ' invalid value
      result=CHR(27, c, a)
    CASE "m" ' console modes a=format b=reserved
      ' format: bit 0-2: 0=standard(4:3) 1=cinema(16:9/16:10/odd) 3=square(1:1) 2=vertical/letter (turning resolution)
      '         4=tiny 8=little 12=low 16=standard 20=cinema(HD) 24=big 31=max (128x128)
      '         only 18 of 31 combinations are valid, see VT 2000 documentation
      '         mainly affects the resolution in characters => how many characters per line,
      '         how many lines; also limited by the canvas resolution, insanely smallest font is 4x8
      IF a=0 OR a>31 THEN RETURN "" ' currently invalid values
      IF b>1 THEN RETURN ""  ' currently unused => should be 0
      result=CHR(27, c, a, 1) ' avoiding Nullbyte trouble
    CASE "P" ' place cursor
      ' only coordinates from 1 to 127 allowed, 0 can mean "don't change"
      IF a>127 OR b>127 THEN RETURN "" ' invalid value
      IF a=0 AND b=0 THEN RETURN "" ' invalid value
      result=CHR(27, c, a, b)
    CASE "p" ' printing colour
      ' #todo: supporting ASCII mode
      IF a>127 OR b>127 THEN
        ' demands UTF-8 encoding
        IF a>127 THEN
          IF b>127 THEN
            result=CHR(27, c, &hC2+INT(a/192), a AND &hBF, &hC2+INT(b/192), b AND &hBF)
          ELSE
            result=CHR(27, c, &hC2+INT(a/192), a AND &hBF, b)
          ENDIF
        ELSE
          result=CHR(27, c, a, &hC2+INT(b/192), b AND &hBF)
        ENDIF
      ELSE
        result=CHR(27, c, a, b)
      ENDIF
    CASE "Q" ' printing quality 1-127 (min-max, <32=draft, 63=default, 95="beautiful")
      IF a=0 OR a>127 THEN RETURN "" ' invalid value
      result=CHR(27, c, a)
    CASE ELSE
      result=CHR(27, c)
  END SELECT

  RETURN result
END FUNCTION


FUNCTION rgRGB(BYVAL red AS INTEGER, green AS INTEGER, blue AS INTEGER) AS UBYTE
  ' the RETROGRA palette is r+g*6+b*36 where 'rgb' values are ranging from 0 to 5
  DIM AS INTEGER i
  DIM AS UBYTE colour

  colour=0
  i=INT((red+11)/20)
  IF i>5 THEN i=5
  IF i<0 THEN i=0
  colour+=i
  i=INT((green+11)/20)
  IF i>5 THEN i=5
  IF i<0 THEN i=0
  i*=6
  colour+=i
  i=INT((blue+11)/20)
  IF i>5 THEN i=5
  IF i<0 THEN i=0
  i*=36
  colour+=i

  IF colour=0 THEN colour=240

  RETURN colour
END FUNCTION


FUNCTION rgRGB(BYVAL hexcolour AS INTEGER) AS UBYTE
  ' the RETROGRA palette is r+g*6+b*36 where 'rgb' values are ranging from 0 to 5
  DIM AS INTEGER i
  DIM AS UBYTE colour

  colour=0
  i=INT((hexcolour+&h190000) / &h330000)
  colour+=i
  i=INT(((hexcolour AND &hFF00)+&h1900) / &h3300)
  i*=6
  colour+=i
  i=INT(((hexcolour AND &hFF)+&h19) / &h33)
  i*=36
  colour+=i

  IF colour=0 THEN colour=240

  RETURN colour
END FUNCTION


SUB rgcSETPIXEL(BYVAL vout AS BYTE PTR, BYVAL colour AS UBYTE)
  DIM AS INTEGER r,g,b
  ' draws a pixel of a console character, performing effects if necessary
  IF colour<240 THEN ' ordinary colour
    *(vout)=colour
  ELSE ' effect colour (0xFx)
    SELECT CASE colour
      CASE 240:
        *vout=0 ' to the console 0xF0 is always black
      CASE 253: ' 0xFD - "relenting" colour, replaced by 0xFE
        *vout=253
      CASE 254: ' 0xFE - transparent colour, only replacing FX-colours
        IF *vout>=240 THEN
          *vout=254
        ENDIF
      CASE 255: ' 0xFF reverse transparent; inverts destination colour (a bit like XOR)
        colour=*vout
        IF colour<216 THEN ' console palette range
          ' invert RGB-based
          colour=215-colour
        ELSE ' user-defined colour, FX-colours are ignored in case
          ' swap colours from 216 to 239
          colour=216+(239-colour)
        ENDIF
      CASE 251: ' 0xFB - darkens ground by 33%
        colour=*vout
        IF colour<216 THEN ' only working with the RGB palette section
          r=colour MOD 6
          g=(colour/6) MOD 6
          b=colour/36
          r-=2
          g-=2
          b-=2
          IF r<0 THEN r=0
          IF g<0 THEN g=0
          IF b<0 THEN b=0
          colour=b*36+g*6+r
          *vout=colour
        ENDIF
      CASE 252: ' 0xFC - lightens ground by 33%
        colour=*vout
        IF colour<216 THEN ' only working with the RGB palette section
          r=colour MOD 6
          g=(colour/6) MOD 6
          b=colour/36
          r+=2
          g+=2
          b+=2
          IF r>5 THEN r=5
          IF g>5 THEN g=5
          IF b>5 THEN b=5
          colour=b*36+g*6+r
          *vout=colour
        ENDIF
      CASE ELSE
        ' preserved effect colours, don't do anything
        ' note: 0xFA (250) stands for "keep previous/current colour", used by ESC p
    END SELECT
  ENDIF
END SUB


SUB rgcREFRESH
  ' puts out the whole console content, based on its char buffer
  DIM AS UBYTE PTR vout, cdp
  DIM AS INTEGER x, y, lo=0, ls, i, j
  DIM AS ULONG char
  DIM AS UBYTE cfl, fgcolour, bgcolour

  vout=SCREENPTR
  lo=retrogra_cwidth
  lo-=8 ' offset to next line (after printing 8 pixels)

  FOR y=0 TO 29
    FOR x=0 TO 79
      IF rgccell(x,y).flags THEN ' character to print
        char=rgccell(x,y).char
        IF rgccell(x,y).flags AND 128 THEN
          ' inverse output
          fgcolour=rgccell(x,y).bgcolour
          bgcolour=rgccell(x,y).fgcolour
        ELSE
          fgcolour=rgccell(x,y).fgcolour
          bgcolour=rgccell(x,y).bgcolour
        ENDIF
        i=char SHR 7
        IF rgcfontpageoffset(i)=0 THEN ' "codepage" not existing => unknown char
          i=0
          char=0
        ENDIF
        ' get char font data start
        ls=rgcfontlinesize(i)
        cdp=rgcfont+rgcfontpageoffset(i)+(char AND 127)
        i=16
        WHILE i
          i-=1
          cfl=PEEK(UBYTE, cdp)
          j=8
          WHILE j
            j-=1
            IF (cfl AND 1) THEN ' foreground pixel
              IF fgcolour<240 THEN
                *vout=fgcolour
              ELSE
                rgcSETPIXEL(vout, fgcolour)
              ENDIF
            ELSE ' background pixel
              IF bgcolour<240 THEN
                *vout=bgcolour
              ELSE
                rgcSETPIXEL(vout, bgcolour)
              ENDIF
            ENDIF
            cfl SHR=1
            vout+=1
          WEND
          cdp+=ls
          vout+=lo
        WEND
        vout-=lo*16+120 ' minus 16 pixel lines on screen & 15*8 => 8 bytes/pixels right
      ELSE
        vout+=8 ' directly 8 pixels right
      ENDIF
    NEXT
    vout+=lo*16-512 ' to begin of next console line in screen memory
  NEXT
END SUB


SUB rgCLS
  ' "Clear Screen" resp. "Clear Home"
  ' empties whole (console) screen, overwriting with current background colour
  ' deleting all characters in the screen buffer
  ' performs a "HOME"(AKA "Pos 1") action for the text cursor
  ' but doesn't change any settings though => may change the screen colour
  DIM AS UBYTE PTR vout
  DIM AS INTEGER x, y, lo=0

  vout=SCREENPTR
  IF retrogra_cwidth>640 THEN lo=retrogra_cwidth-640

  ' set console area to background colour
  IF rgcbgcolour<240 THEN
    FOR y=0 TO 479
      FOR x=0 TO 639
        *vout=rgcbgcolour
        vout+=1
      NEXT
      vout+=lo
    NEXT
  ELSE ' background colour is special effect colour
    FOR y=0 TO 479
      FOR x=0 TO 639
        rgcSETPIXEL(vout, rgcbgcolour)
        vout+=1
      NEXT
      vout+=lo
    NEXT
  ENDIF
  FOR y=0 TO 29
    FOR x=0 TO 79
      rgccell(x,y).flags=0
      IF rgcfgcolour=240 THEN
        rgccell(x,y).fgcolour=0 ' "permanent black"
      ELSE
        rgccell(x,y).fgcolour=rgcfgcolour
      ENDIF
      IF rgcbgcolour=240 THEN
        rgccell(x,y).bgcolour=0 ' "permanent black"
      ELSE
        rgccell(x,y).bgcolour=rgcbgcolour
      ENDIF
      rgccell(x,y).char=0
    NEXT
  NEXT
END SUB


SUB rgcSCROLLUP(BYVAL lines AS INTEGER=1)
  DIM AS INTEGER x, y

  IF lines<2 THEN ' =>scroll 1 line
    ' copy cells one line higher
    FOR y=0 TO 28
      FOR x=0 TO 79
        rgccell(x,y)=rgccell(x, y+1)
      NEXT
    NEXT
    ' empty last(new) line with spaces
    FOR x=0 TO 79
      rgccell(x,29).flags=1
      rgccell(x,29).fgcolour=rgcfgcolour
      rgccell(x,29).bgcolour=rgcbgcolour
      rgccell(x,29).char=32
    NEXT
  ELSE
    IF lines<29 THEN
      ' copy cells by <lines> line higher
      FOR y=0 TO 29-lines
        FOR x=0 TO 79
          rgccell(x,y)=rgccell(x, y+lines)
        NEXT
      NEXT
      ' empty last(new) lines with spaces
      FOR y=lines TO 29
        FOR x=0 TO 79
          rgccell(x,y).flags=1
          rgccell(x,y).fgcolour=rgcfgcolour
          rgccell(x,y).bgcolour=rgcbgcolour
          rgccell(x,y).char=32
        NEXT
      NEXT
    ELSE ' for the joker who wants to "scroll" more lines than a screen has
      rgCLS
      EXIT SUB
    ENDIF
  ENDIF
  rgcREFRESH
END SUB


SUB rgcSCROLLDOWN(BYVAL lines AS INTEGER=1)
  DIM AS INTEGER x, y

  IF lines<2 THEN ' =>scroll 1 line
    ' copy cells one line down
    FOR y=28 TO 0 STEP -1
      FOR x=0 TO 79
        rgccell(x,y+1)=rgccell(x, y)
      NEXT
    NEXT
    ' empty first(new) line with spaces
    FOR x=0 TO 79
      rgccell(x,0).flags=1
      rgccell(x,0).fgcolour=rgcfgcolour
      rgccell(x,0).bgcolour=rgcbgcolour
      rgccell(x,0).char=32
    NEXT
  ELSE
    IF lines<29 THEN
      ' copy cells by <lines> line higher
      FOR y=29-lines TO 0 STEP -1
        FOR x=0 TO 79
          rgccell(x,y+lines)=rgccell(x, y)
        NEXT
      NEXT
      ' empty first(new) lines with spaces
      FOR y=0 TO lines-1
        FOR x=0 TO 79
          rgccell(x,y).flags=1
          rgccell(x,y).fgcolour=rgcfgcolour
          rgccell(x,y).bgcolour=rgcbgcolour
          rgccell(x,y).char=32
        NEXT
      NEXT
    ELSE ' for the joker who wants to "scroll" more lines than a screen has
      RGCLS
      EXIT SUB
    ENDIF
  ENDIF
  rgcREFRESH
END SUB


SUB rgcBELL
  ' bell sound (alert sound)
  ' #todo:
  BEEP
END SUB


SUB rgcTYPE
  ' type sound
  ' #todo:
END SUB


SUB rgcCARRIAGERETURN
  ' carriage return sound
  ' #todo:
END SUB


SUB rgcMOVEON(BYVAL cells AS INTEGER=1)
  ' moves console "cursor" 1 space (after putting out a char mainly)
  ' #todo: maybe 4 cells for double-sized emphasized characters (or maybe stupid)
  IF cells<1 OR cells>2 THEN EXIT SUB ' illegal value
  rgccurx+=cells
  ' line overflow check
  IF rgccurx>80 THEN ' more 2 or more chars over the limit
    IF (rgcflags AND 2) THEN
      ' line overflow (to beginning of next line)
      rgccurx-=80
      rgccury+=1
      IF rgccury>29 THEN
        IF (rgcflags AND 1) THEN ' scrolling enabled
          RGCSCROLLUP
        ENDIF
        rgccury=29 ' scrolling disabled - trapped in last line
      ENDIF
    ELSE ' line overflow is off
      IF (rgcflags AND 128) THEN
        rgcBELL
      ENDIF
      rgccurx=79
    ENDIF
  ENDIF
END SUB


SUB rgcPRINTCHAR(BYVAL char AS ULONG) ' puts out given char (graphic, without control code handling)
  DIM AS UBYTE PTR cd, vout
  DIM AS INTEGER i, x, y, cx, cy, scw, ls
  DIM AS UBYTE cl, fgcolour

  ' scrolling is only performed when another character is put out
  IF rgccurx>79 THEN
    IF (rgcflags AND 2) THEN
      ' line overflow (to beginning of next line)
      rgccurx-=80
      rgccury+=1
      IF rgccury>29 THEN
        IF (rgcflags AND 1) THEN ' scrolling enabled
          rgcSCROLLUP
        ENDIF
        rgccury=29
      ENDIF
    ELSE ' line overflow is off
      IF (rgcflags AND 128) THEN
        rgcBELL
      ENDIF
      rgccurx=79
    ENDIF
  ENDIF

  vout=SCREENPTR ' error source: should be currently in the virtual 640x480 screen
  scw=retrogra_cwidth

  i=char SHR 7 ' "character page"
  IF rgcfontpageoffset(i) THEN
    cd=rgcfont+rgcfontpageoffset(i)+(char AND 127) ' #note: this only works for 1 byte broad chars
    ls=rgcfontlinesize(i)
  ELSE ' offset is 0 => "page" not existing, put out "unknown" char (code 0)
    cd=rgcfont+rgcfontpageoffset(0) ' #note: "codepage" 0 must always exist
    ls=rgcfontlinesize(0)
    char=0
  ENDIF

  ' draw character
  vout+=rgccury*scw*16+rgccurx*8
  scw-=8
  IF rgccharflags AND 2 THEN ' bold output
    ' little cheating: make printing colour darker or brighter to get
    ' a bold printing effect
    IF rgcfgcolour=0 OR rgcfgcolour>214 THEN
      ' black, white, userdefined or effect colour => nothing possible
      fgcolour=rgcfgcolour
    ELSE
      DIM AS UBYTE r, g, b
      DIM AS BOOLEAN darkground=TRUE

      IF rgcbgcolour>0 AND rgcbgcolour<216 THEN
        ' BG not black or effect colour
        IF rgcbgcolour>=144 THEN
          darkground=FALSE
        ELSE
          r=rgcbgcolour MOD 6
          g=INT((rgcbgcolour MOD 36)/6)
          IF r>2 OR g>1 THEN darkground=FALSE
        ENDIF
      ENDIF
      r=rgcfgcolour MOD 6
      g=INT((rgcfgcolour MOD 36)/6)
      b=INT(rgcfgcolour/36)
      IF (darkground<>0) AND (r>2 OR g>1) THEN
        ' bright colour on dark ground => make brighter
        IF (r=5) OR (g=5) OR (b=5) THEN
          ' at least one RGB value is maximum => impossible
          fgcolour=rgcfgcolour
        ELSE
          r+=1
          g+=1
          b+=1
          fgcolour=r+g*6+b*36
        ENDIF
      ELSE
        ' dark colour / light ground => make darker
        ' preferring 2 steps due to bad contrast
        IF r>1 THEN
          r-=2
        ELSEIF r THEN r-=1
        ENDIF
        IF g>1 THEN
          g-=2
        ELSEIF g THEN g-=1
        ENDIF
        IF b>1 THEN
          b-=2
        ELSEIF b THEN b-=1
        ENDIF
        fgcolour=r+g*6+b*36
      ENDIF
    ENDIF
  ELSE
    fgcolour=rgcfgcolour
  ENDIF
  IF rgccharflags AND 128 THEN ' inverted output
    IF rgcfgcolour>239 OR rgcbgcolour>240 THEN ' using effect colours
      ' slower like with normal colours
      FOR y=0 TO 15
        cl=PEEK(UBYTE, cd)
        FOR x=0 TO 7
          IF (cl AND 1) THEN
            rgcSETPIXEL(vout, rgcbgcolour)
          ELSE
            rgcSETPIXEL(vout, rgcfgcolour)
          ENDIF
          cl SHR=1
          vout+=1
        NEXT
        cd+=ls
        vout+=scw
      NEXT
    ELSE ' normal colours, replacing background
      FOR y=0 TO 15
        cl=*(cd)
        FOR x=0 TO 7
          IF (cl AND 1) THEN
            *(vout)=rgcbgcolour
          ELSE
            *(vout)=fgcolour
          ENDIF
          cl SHR=1
          vout+=1
        NEXT
        cd+=ls
        vout+=scw
      NEXT
    ENDIF
  ELSE ' non-inverted output
    IF rgcfgcolour>239 OR rgcbgcolour>240 THEN ' using effect colours
      ' slower like with normal colours
      FOR y=0 TO 15
        cl=*(cd)
        FOR x=0 TO 7
          IF (cl AND 1) THEN
            rgcSETPIXEL(vout, rgcfgcolour)
          ELSE
            rgcSETPIXEL(vout, rgcbgcolour)
          ENDIF
          cl SHR=1
          vout+=1
        NEXT
        cd+=ls
        vout+=scw
      NEXT
    ELSE ' normal colours, replacing background
      FOR y=0 TO 15
        cl=*(cd)
        FOR x=0 TO 7
          IF (cl AND 1) THEN
            *(vout)=fgcolour
          ELSE
            *(vout)=rgcbgcolour
          ENDIF
          cl SHR=1
          vout+=1
        NEXT
        cd+=ls
        vout+=scw
      NEXT
    ENDIF
  ENDIF

  ' store char in console buffer
  ' #todo: 2 cell chars, check...
  rgccell(rgccurx,rgccury).flags=rgccharflags
  rgccell(rgccurx,rgccury).fgcolour=rgcfgcolour
  rgccell(rgccurx,rgccury).bgcolour=rgcbgcolour
  rgccell(rgccurx,rgccury).char=char

  ' move 1 char #todo: double size characters => 2 cells
  rgcMOVEON
END SUB


SUB rgPRINT(BYVAL outputstring AS STRING)
  DIM AS INTEGER strend=LEN(outputstring), i, j
  DIM AS UBYTE c
  DIM AS ULONG uc

  IF strend=0 THEN EXIT SUB ' empty string
  strend-=1
  FOR i=0 TO strend
    c=PEEK(UBYTE, STRPTR(outputstring)+i) ' maybe faster than MID$()...?
    IF c<32 THEN ' control code
      IF c THEN ' not 0
        IF c=27 THEN ' ESC (sequence)
          IF i<strend THEN
            i+=1
            c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
            IF c<128 THEN ' 7-bit / ASCII range
              SELECT CASE c
                CASE &H41: ' ESC+A inverse mode on
                  rgccharflags OR=128
                CASE &H61: ' ESC+a inverse mode off
                  rgccharflags AND=NOT 128
                CASE &H42: ' ESC+B bold on
                  rgccharflags OR=2
                CASE &H62: ' ESC+b bold off
                  rgccharflags AND=NOT 2
                CASE &H43: ' ESC+C canceled on
                  rgccharflags OR=16
                CASE &H63: ' ESC+c canceled off
                  rgccharflags AND=NOT 16
                CASE &H44: ' ESC+D disabled on
                  rgccharflags OR=32
                CASE &H64: ' ESC+d disabled off
                  rgccharflags AND=NOT 32
                CASE &H45: ' ESC+E emphasized/double width on
                  rgccharflags OR=1024
                CASE &H65: ' ESC+e normal width
                  rgccharflags AND=NOT 1024
                CASE &H46: ' ESC+F compressed mode on
                  rgccharflags OR=4096
                CASE &H47: ' ESC+G <bitmap data> graphic output (unsupported)
                CASE &H48: ' ESC+H double height on
                  rgccharflags OR=2048
                CASE &H68: ' ESC+h double height off
                  rgccharflags AND=NOT 2048
                CASE &H49: ' ESC+I italic on
                  rgccharflags OR=4
                CASE &H69: ' ESC+i italic off
                  rgccharflags AND=NOT 4
                CASE &H4A: ' ESC+J Form Feed / next page
                  ' on a screen, it's just like CLS - but we put a white paper in our printer ;-)...
                  rgcfgcolour=0
                  rgcbgcolour=215
                  RGCLS
                CASE &H6A: ' ESC+j previous page
                  ' also nonsense for a screen
                  rgCLS
                CASE &H4B: ' ESC+K erase rest of screen
                CASE &H6B: ' ESC+k erase rest of line
                CASE &H4D: ' ESC+M <c> "select codepage" (unsupported)
                  ' #todo: strip off character 'c' from string
                CASE &H4E: ' ESC+N all decorations&scalings off => normal text
                  rgccharflags=1
                CASE &H6E: ' ESC+n normal text size (Esc E/H/S/s off)
                  rgccharflags AND=NOT 7936
                CASE &H4F: ' ESC+O outlined on
                  rgccharflags OR=64
                CASE &H6F: ' ESC+o outlined off
                  rgccharflags AND=NOT 64
                CASE &H50: ' ESC+P place cursor <x> <y>
                  ' might use '127' to not change the coordinate, but "VT-2000" supports
                  ' in principle a maximum page size of 128x128 - from 1024x2048 up, overall
                  ' 3 page sizes with 128 characters per line - which is even not
                  ' entirely covered by ESC P
                  ' x and y must not be 0 nor above 127 though, they start with (1|1)
                  ' #note: internally, the coordinates are reaching from 0 to 79/29
                  IF (i+2)<=strend THEN ' not yet string end reached
                    i+=1
                    c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
                    IF c AND c<81 THEN ' other values are illegal
                      rgccurx=c-1
                    ENDIF
                    i+=1
                    c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
                    IF c AND c<31 THEN ' other values are illegal
                      rgccury=c-1
                    ENDIF
                  ELSE
                    EXIT SUB
                  ENDIF
                CASE &H70: ' ESC+p set printing colours <fg> <bg>
                  ' => foreground & background colour
                  ' because the valid values range from 1 to 255, they may be also
                  ' 2 byte long UTF-8 encoded
                  ' 0 is illegal, you have to use 0xF0 instead chr(195,176)
                  IF (i+1)<=strend THEN ' string not too short
                    ' foreground colour
                    i+=1
                    c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
                    IF c THEN ' Nullbyte is illegal => you have to use 0xF0 / 240 CHR(195,176)
                      IF c<128 THEN ' colour below 128, easy
                        rgcfgcolour=c
                      ELSE ' UTF-8 encoded colour
                        IF c=194 THEN ' only those 2 values are valid
                          uc=128
                        ELSEIF c=195 THEN
                          uc=192
                        ELSE ' not ASCII nor UTF-8 (resp. valid)
                          EXIT SUB
                        ENDIF
                        i+=1
                        c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
                        uc OR=(c AND 63)
                        IF uc<>250 THEN rgcfgcolour=uc ' colour 0xFA - don't change colour
                      ENDIF
                    ENDIF
                    i+=1
                    IF i>strend THEN EXIT SUB ' end of string
                    c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
                    ' background colour
                    IF c THEN ' Nullbyte is illegal => you have to use 0xF0 / 240 CHR(195,176)
                      IF c<128 THEN ' colour below 128, easy
                        rgcbgcolour=c
                      ELSE ' UTF-8 encoded colour
                        IF c=194 THEN ' only those 2 values are valid
                          uc=128
                        ELSEIF c=195 THEN
                          uc=192
                        ELSE ' not ASCII nor UTF-8 (resp. valid)
                          EXIT SUB
                        ENDIF
                        i+=1
                        IF i>strend THEN EXIT SUB ' end of string
                        c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
                        uc OR=(c AND 63)
                        IF uc<>250 THEN rgcbgcolour=uc ' colour 0xFA - don't change colour
                      ENDIF
                    ENDIF
                  ENDIF
                CASE &H51: ' ESC+Q <c> printing quality 0-127 (0 draft, 31 fast, 32 normal, 64+ highest)
                  ' #todo: strip off character 'c' from string
                CASE &H52: ' ESC+R proportional mode (reserved for)
                CASE &H72: ' ESC+r monospaced mode (default)
                CASE &H53: ' ESC+S superscript on
                  rgccharflags OR=512
                CASE &H73: ' ESC+s subscript on
                  rgccharflags OR=256
                CASE &H55: ' ESC+U underlined on
                  rgccharflags OR=8
                CASE &H75: ' ESC+U underlined off
                  rgccharflags AND=NOT 8
                CASE &H58: ' ESC+X show cursor (always)
                  rgcflags OR=32
                CASE &H78: ' ESC+x hide cursor
                  rgcflags AND=NOT 32
                CASE &H59: ' ESC+Y line overflow on
                  rgcflags OR=2
                CASE &H79: ' ESC+y line overflow off
                  rgcflags AND=NOT 2
                CASE &H5A: ' ESC+Z scrolling on
                  rgcflags OR=1
                CASE &H7A: ' ESC+z scrolling off
                  rgcflags AND=NOT 1
                CASE ELSE:
                  ' ignore illegal/unsupported ESC sequences, but put out remaining chars
                  i-=1
              END SELECT
            ELSE
              i-=1 ' illegal value => ignore "Mask" code, proceed
            ENDIF
          ELSE
            EXIT SUB
          ENDIF
        ELSE
          SELECT CASE c ' sorted by relevance
            CASE 13: ' return (+scrolling)
              rgccurx=0
              rgccury+=1
              IF rgccury>29 THEN
                rgccury=29
                IF (rgcflags AND 1) THEN rgcSCROLLUP ' scrolling enabled
              ENDIF
            CASE 8: ' 8 tab right (ordinary tabulator, 8 char steps)
              IF rgccurx>78 THEN ' last column => new line
                rgccurx=79
                rgcMOVEON
              ELSE
                rgccurx+=8
                rgccurx AND=120
              ENDIF
            CASE 1: ' cursor up (+scrolling)
              IF rgccury THEN
                rgccury-=1
              ELSE
                IF (rgcflags AND 1) THEN rgcSCROLLDOWN
              ENDIF
            CASE 2: ' cursor down (+scrolling) "Line Feed" LF
              rgccury+=1
              IF rgccury>29 THEN
                rgccury=29
                rgcSCROLLUP
              ENDIF
            CASE 3: ' cursor left
              IF rgccurx THEN
                rgccurx-=1
              ELSE
                rgccurx=79
                IF rgccury THEN
                  rgccury-=1
                ELSE
                  rgcSCROLLDOWN
                ENDIF
              ENDIF
            CASE 4: ' cursor right
              IF rgccurx>78 THEN
                rgccurx=0
              ELSE
                rgccurx=79
                IF rgccury THEN
                  rgccury-=1
                ELSE
                  rgcSCROLLDOWN
                ENDIF
              ENDIF
            CASE 26: ' Mask
              ' puts out next char's graphic minus bit 5, e.g. a control code graphic
              IF i<strend THEN
                i+=1
                c=PEEK(UBYTE, STRPTR(outputstring)+i) ' get next char
                IF c<128 THEN ' 7-bit / ASCII range
                  c AND=95
                  rgcPRINTCHAR(CAST(ULONG,c))
                ELSE
                  ' #todo: UTF-8 support, putting out normally invisible control codes
                  i-=1 ' illegal value => ignore "Mask" code, proceed
                ENDIF
              ELSE
                EXIT SUB
              ENDIF
            CASE 24: ' 4 tab right
              IF rgccurx>78 THEN ' last column => new line
                rgcMOVEON
              ELSE
                rgccurx+=4
                rgccurx AND=124
              ENDIF
            CASE 14: ' clear line CL (+ code 11)
            CASE 15: ' clear screen CLS (+ HOME = code 11+9 = code 16)
              rgCLS
            CASE 16: ' HOME ("Pos 1")
              rgccurx=0
              rgccury=0
            CASE 19: ' Backspace
            CASE 20: ' Delete
            CASE 25: ' Insert
            CASE 28: ' redraw screen
            CASE 23: ' 4 tab left (previous tab position)
              IF rgccurx THEN
                rgccurx-=1
                rgccurx AND=124
              ELSE
                rgccurx=79
                IF rgccury THEN
                  rgccury-=1
                ELSE
                  rgcSCROLLDOWN
                ENDIF
              ENDIF
            CASE 7: ' 8 tab left (previous tab position)
              IF rgccurx THEN
                rgccurx-=1
                rgccurx AND=120
              ELSE
                rgccurx=79
                IF rgccury THEN
                  rgccury-=1
                ELSE
                  rgcSCROLLDOWN
                ENDIF
              ENDIF
            CASE 9: ' cursor to top
              rgccury=0
            CASE 10: ' cursor to bottom
              rgccury=29
            CASE 11: ' cursor to left border / "Carriage Return" CR
              rgccurx=0
            CASE 12: ' cursor to right border
              rgccurx=79
            CASE 29: ' carriage returning sound (scrolling)
              rgcCARRIAGERETURN
            CASE 30: ' typing sound
              rgcTYPE
            CASE 31: ' Bell / Alert, End Of Line (right paper edge) signal/sound
              rgcBELL
            CASE ELSE: ' unsupported & illegal control codes => ignored
              ' 5, 6, 21, 22 vertical "tabulators" => useless
          END SELECT
        ENDIF
      ELSE ' c is 0
        EXIT SUB ' Nullbyte => break output
      ENDIF
    ELSE ' normal character
      IF c>127 THEN ' UTF-8 char
        IF (c AND 192)=192 THEN ' UTF-8 start character
          IF (c AND 224)=192 THEN ' "110..." => 2 byte character
            uc=(c AND 63) SHL 6
            i+=1
            c=PEEK(UBYTE, STRPTR(outputstring)+i)
            IF (c AND 192)=128 THEN ' okay
              uc+=(c AND 63)
              ' characters from 127 to 159 are invisible control characters
              ' 173 is invisible Soft Hyphen (SHY)
              IF uc>159 AND uc<>173 THEN rgcPRINTCHAR(uc)
            ELSE
              EXIT SUB ' illegal UTF-8 code #todo:
            ENDIF
          ELSEIF (c AND 240)=224 THEN ' "1110..." => 3 byte character
            uc=(c AND 15) SHL 12
            i+=1
            c=PEEK(UBYTE, STRPTR(outputstring)+i)
            IF (c AND 192)=128 THEN ' okay
              uc+=(c AND 63) SHL 6
              i+=1
              c=PEEK(UBYTE, STRPTR(outputstring)+i)
              IF (c AND 192)=128 THEN ' okay
                uc+=(c AND 63)
                rgcPRINTCHAR(uc)
              ELSE
                EXIT SUB ' illegal UTF-8 code #todo:
              ENDIF
            ELSE
              EXIT SUB ' illegal UTF-8 code #todo:
            ENDIF
          ELSEIF (c AND 248)=240 THEN ' "11110..." => 4 byte character
            uc=(c AND 63) SHL 6
          ELSE ' 5 byte character or more => illegal
            EXIT SUB ' #todo: maybe another handling / reaction; error code
          ENDIF
        ELSE
          ' not a UTF-8 char #todo: ? - could be only some binary data or ASCII char, but...
        ENDIF
      ELSE ' ASCII char
        IF c<127 THEN ' 127 is ASCII-Delete (invisible)
          uc=c
          rgcPRINTCHAR(uc)
        ENDIF
      ENDIF
    ENDIF
  NEXT
END SUB


#DEFINE rgINKEY INKEY8
' (alias, might become official)
FUNCTION INKEY8() AS STRING
  ' this function should guarant a platform independend INKEY, returning UTF-8 encoded characters
  DIM AS STRING result
  DIM AS UBYTE c
  DIM AS USHORT uc
  ' #todo: Linux support / scancodes / testing
  ' #todo: maybe use native code instead without using INKEY() and MULTIKEY()
  ' #todo: #note: only for Windows with Western European encoding (is DOS CP 850)
  '        needs support for Linux, and other Windows encodings
  DIM AS USHORT ucode(128)={ _
  &hC387, &hC3BC, &hC3A9, &hC3A2, &hC3A4, &hC3A0, &hC3A5, &hC3A7, &hC3AA, &hC3AB, &hC3A8, &hC3AF, &hC3AE, &hC3AC, &hC384, &hC385, _
  &hC389, &hC3A6, &hC386, &hC3B4, &hC3B6, &hC3B2, &hC3BB, &hC3B9, &hC3BF, &hC396, &hC39C, &hC3B8, &hC2A3, &hC398, &hC397, &hC692, _
  &hC3A1, &hC3AD, &hC3B3, &hC3BA, &hC3B1, &hC391, &hC2AA, &hC2BA, &hC2BF, &hC2AE, &hC2AC, &hC2BD, &hC2BC, &hC2A1, &hC2AB, &hC2BB, _
  &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
  &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, _
  &hC3B0, &hC390, &hC38A, &hC38B, &hC388, &h0000, &hC38D, &hC38E, &hC38F, &h0000, &h0000, &hC2A6, &h0000, &hC2A6, &hC38C, &hC2AF, _
  &hC393, &hC39F, &hC394, &hC392, &hC3B5, &hC395, &hC2B5, &hC3BE, &hC39E, &hC39A, &hC39B, &hC399, &hC3BD, &hC39D, &hC2AF, &hC2B4, _
  &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000, &h0000 }

  ' determine shift key status
  rgcshiftkeys=NONE
  IF (MULTIKEY(FB.SC_LSHIFT) OR MULTIKEY(FB.SC_RSHIFT)) THEN ' Shift
    rgcshiftkeys OR=SHIFT
  ENDIF
  IF MULTIKEY(FB.SC_ALT) THEN ' Alternate
    rgcshiftkeys OR=ALT
  ENDIF
  IF MULTIKEY(FB.SC_CONTROL) THEN ' Control (Command)
    rgcshiftkeys OR=CMD
  ENDIF
  IF MULTIKEY(FB.SC_CAPSLOCK) THEN ' Caps Lock (Shift Lock)
    rgcshiftkeys OR=CAPSLOCK
  ENDIF

  result=INKEY

  IF result="" THEN
    #IFDEF __FB_WIN32__
    ' determine whether F10 is pressed (Windows bug)
    IF MULTIKEY(FB.SC_F10) THEN
      SLEEP 25 ' bad workaround to avoid
      RETURN CHR(249) ' code for F10
    ELSE
      RETURN "" ' keyboard buffer is empty
    ENDIF
    #ELSE
    RETURN "" ' keyboard buffer is empty
    #ENDIF
  ENDIF

  IF LEN(result)>1 THEN ' special keys
    IF result[0]=255 THEN ' really a special key's code
      c=result[1]
      SELECT CASE AS CONST c
        CASE 72: ' Cursor up
          RETURN CHR(1)
        CASE 141: ' (Ctrl+) Cursor up
          RETURN CHR(1)
        CASE 80: ' Cursor down
          RETURN CHR(2)
        CASE 145: ' (Ctrl+) Cursor down
          RETURN CHR(2)
        CASE 75: ' Cursor left
          RETURN CHR(3)
        CASE 115: ' (Ctrl+) Cursor left
          RETURN CHR(3)
        CASE 77: ' Cursor right
          RETURN CHR(4)
        CASE 116: ' (Ctrl+) Cursor right
          RETURN CHR(4)
        CASE 82: ' Insert (<-> Overwrite)
          RETURN CHR(25)
        CASE 146: ' (Ctrl+) Insert (<-> Overwrite)
          RETURN CHR(25)
        CASE 83: ' Delete
          RETURN CHR(20)
        CASE 147: ' (Ctrl+) Delete
          RETURN CHR(20)
        CASE 71: ' (Clear) Home
          RETURN CHR(16)
        CASE 119: ' (Ctrl+) (Clear) Home
          RETURN CHR(16)
        CASE 79: ' End
          RETURN CHR(17)
        CASE 117: ' (Ctrl+) End
          RETURN CHR(17)
        CASE 73: ' Page up
          RETURN CHR(9)
        CASE 132: ' (Ctrl+) Page up
          RETURN CHR(9)
        CASE 81: ' Page down
          RETURN CHR(10)
        CASE 118: ' (Ctrl+) Page down
          RETURN CHR(10) ' 3B 54 5E   59 84 94
        CASE 59 TO 68: ' F1-F10 => &hF0-&hFF
          RETURN CHR(c+181)
        CASE 84 TO 93: ' (Shift+) F1-F10 => &hF0-&hFF
          RETURN CHR(c+156)
        CASE 94 TO 103: ' (Ctrl+) F1-F10 => &hF0-&hFF
          RETURN CHR(c+146)
        CASE 104 TO 113: ' (AltGr+) F1-F10 => &hF0-&hFF
          RETURN CHR(c+136)
        CASE 133 TO 140 ' F11+F12 => &hFA+&hFB
          ' several different scancodes, all being F11 or F12 combined with Shift, Control, Alt...
          RETURN CHR(&hFB-(c MOD 2))
          ' #note: scancodes for F13-F16 ? (most likely not supported by FreeBASIC)
      END SELECT
    ENDIF
    RETURN ""
  ELSE
    c=result[0]
    IF c>31 AND c<127 THEN ' ASCII range, 1:1 (128 is normally Delete)
      RETURN result
    ELSEIF c>127 THEN ' a case for Unicode/UTF-8
      uc=ucode(c-128) ' get Unicode from the character table
      IF uc THEN ' return UTF-8 sequence - guaranted 2 bytes long (up to U+07FF)
        result=CHR(uc SHR 8, uc AND 255)
        RETURN result
      ELSE
        RETURN "" ' character not supported
      ENDIF
    ELSE
      IF c=13 THEN ' Return
        RETURN CHR(13)
      ELSEIF c=9 THEN ' Tabulator
        RETURN CHR(8)
      ELSEIF c=8 THEN ' Backspace
        RETURN CHR(19)
      ELSEIF c=27 THEN ' Escape
        RETURN CHR(27)
      ENDIF
      RETURN "" ' unsupported code
    ENDIF
  ENDIF

  RETURN result
END FUNCTION

FUNCTION rgSHIFTKEYS(mask AS rgSHIFTKEY=255) AS rgSHIFTKEY
  ' will return all flags except "lock" keys
  RETURN (rgcshiftkeys AND mask)
END FUNCTION


SUB rgINPUT(BYVAL prompt AS STRING, BYREF inputstr AS STRING, BYVAL limit AS INTEGER=72, BYVAL areawidth AS INTEGER=0)
  DIM AS INTEGER ibx, iby, ibw, its ' input box x-pos., y-pos. & max width, input string start pos
  DIM AS INTEGER mx, my, mw, mb ' mouse state variables
  DIM AS INTEGER cpos ' cursor position
  DIM AS INTEGER i ' dummy/multipurpose integer
  DIM AS STRING key
  DIM AS UBYTE c
  DIM AS rgMOUSEFORM mouseform

  ' #note: this version only works for RETROGRA, it expects 80x30 characters in a 640x480 resolution
  ' #todo: mouse placing of cursor and text selection
  ' #todo: shift-selection of text
  ' #todo: copy, cut & paste, the best using the OS clipboard

  IF prompt<>"" THEN rgPRINT prompt ' put out prompt

  ibx=rgccurx+1 ' get current cursor position (=> input area start)
  iby=rgccury+1
  ' determine input area size
  IF areawidth THEN
    IF areawidth<3 THEN ' ridiculous and causes errors
      inputstr=""
      RETURN
    ENDIF
    IF ibx+areawidth>80 THEN
      ibw=81-ibx
    ELSE
      ibw=areawidth
    ENDIF
  ELSE
    ibw=81-ibx
  ENDIF

  IF inputstr="" THEN ' input string not preallocated
    its=1
    cpos=1
  ELSE ' input string has a preallocation
    ' place cursor behind string
    cpos=LEN(inputstr)+1
    IF UTF8LEN(inputstr)>ibw THEN ' preallocation is longer than input area
      its=UTF8GETPOS(inputstr, LEN(inputstr), 2-ibw)
      ' (minus arrow char at the end; an offset of -1 means 2 characters)
    ELSE ' full input text from the prompt on
      its=1
    ENDIF
  ENDIF

  IF retrogramousepointer<>0 THEN
    ' => selections with mouse
    mouseform=retrogramousepointer
    rgSETMOUSE BRACKET
  ENDIF

  IF its=1 THEN
    rgPRINT inputstr+CHR(27,65,32,27,97)
  ELSE
    rgPRINT CHR(26,35)+UTF8MID(inputstr,its,1)+CHR(27,65,32,27,97) ' with left arrow
  ENDIF
  ' empty keyboard buffer
  WHILE INKEY8()<>""
  WEND
  DO
    key=INKEY8()
    IF retrogramousepointer<>0 THEN
      rgGETMOUSE mx, my, mw, mb
      IF mb<>0 THEN
        mx/=8
        my/=16
        IF (mx+8)<ibx OR (mx-8)>(ibx+ibw) THEN EXIT DO ' click far outside => Escape
        IF (my+4)<iby OR (my-4)>iby THEN EXIT DO ' click far outside => Escape
      ENDIF
    ENDIF

    IF key<>"" THEN
      SELECT CASE key
        CASE CHR(27)
          EXIT DO
        CASE CHR(13)
          EXIT DO
        CASE CHR(3) ' left
          ' #todo: test for SHIFT (select)
          IF cpos>1 THEN
            cpos=UTF8GETPOS(inputstr, cpos, -1) ' one char back
            IF rgSHIFTKEYS(SHIFT+CMD)=CMD THEN ' Control/Command pressed => word jumping
              ' get last space/word beginning
              cpos=UTF8GETPOS(inputstr, cpos, -1) ' one char back
              IF cpos=0 THEN cpos=1 ' (error), to string start
              cpos=INSTRREV(inputstr, " ", cpos) ' goto last space
              IF cpos<2 THEN ' line beginning reached
                cpos=1
              ELSE
                cpos=UTF8GETPOS(inputstr, cpos, 1) ' goto 1st char of word (after Space)
              ENDIF
            ENDIF
            IF cpos=0 THEN cpos=1 ' (error), to string start
          ENDIF
        CASE CHR(4) ' right
          ' #todo: test for SHIFT (select)
          cpos=UTF8GETPOS(inputstr, cpos, 1) ' next char's position
          IF cpos=0 THEN cpos=LEN(inputstr)+1 ' (error), to end of input
          IF rgSHIFTKEYS(SHIFT+CMD)=CMD THEN ' Control/Command pressed => word jumping
            ' get last space/word beginning
            cpos=INSTR(cpos, inputstr, " ") ' goto last space
            IF cpos THEN
              cpos=UTF8GETPOS(inputstr, cpos, 1) ' goto 1st char of word (after Space)
            ELSE
              cpos=LEN(inputstr)+1 ' to end of string
            ENDIF
            IF cpos=0 THEN cpos=LEN(inputstr)+1 ' (maybe error), to end of string
          ENDIF
        CASE CHR(16) ' Home
          cpos=1
        CASE CHR(17) ' End
          cpos=LEN(inputstr)+1
        CASE CHR(19) ' Backspace
          IF cpos>1 THEN ' = inputstring not empty / deletable character
            cpos=UTF8GETPOS(inputstr, cpos, -1)
            IF cpos=1 THEN ' delete first char
              IF UTF8LEN(inputstr)>1 THEN ' inputstring is longer than 1 char
                inputstr=UTF8MID(inputstr, 1, 1) ' from 1 byte +1 char on
              ELSE
                inputstr="" ' 1 char => empty string
              ENDIF
            ELSE ' delete middle char
              inputstr=UTF8LEFT(inputstr, cpos, -1)+UTF8MID(inputstr, cpos, 1)
            ENDIF
          ENDIF
        CASE CHR(20) ' Delete
          IF cpos<=LEN(inputstr) THEN ' cursor on or before last char
            inputstr=UTF8LEFT(inputstr, cpos, -1)+UTF8MID(inputstr, cpos, 1)
          ENDIF          
        CASE ELSE
          IF LEN(key)=1 AND ASC(key)<32 THEN
            ' unused/unsupported control code, ignore
          ELSEIF LEN(key)=1 AND ASC(key)>239 THEN
            ' function key, ignore
          ELSE
            ' character => add
            IF UTF8LEN(inputstr)+UTF8LEN(key)<=limit THEN
              IF cpos>=UTF8GETPOS(inputstr,LEN(inputstr)) THEN ' cpos is last char
                inputstr+=key
              ELSE
                inputstr=UTF8LEFT(inputstr,cpos,-1)+key+UTF8MID(inputstr,cpos,0)
              ENDIF
              cpos+=LEN(key)
            ENDIF
          ENDIF
      END SELECT

      rgPRINT AT(ibx, iby)+SPACE(ibw) ' clear "input field"
      IF UTF8LEN(inputstr)>=ibw THEN ' input string has to be limited
        ' determine values / scroll string if necessary
        IF cpos>its THEN ' maybe scroll left
          i=UTF8GETDIFF(inputstr,its,cpos)
          IF i>=ibw-1 THEN ' (cpos-its) is bigger than input field
            its=UTF8GETPOS(inputstr,cpos,2-ibw)
          ENDIF
        ELSEIF cpos<=its THEN ' have to scroll right
          IF cpos=1 THEN
            its=1
          ELSE
            its=UTF8GETPOS(inputstr,cpos,-1)
          ENDIF
        ENDIF

        ' put out input string (and cursor)
        rgPRINT AT(ibx, iby)
        IF cpos>LEN(inputstr) THEN
          ' put cursor after input string
          its=UTF8GETPOS(inputstr,LEN(inputstr),2-ibw)
          rgPRINT UTF8MID(inputstr,its,0)+CHR(27,65,32,27,97)
        ELSEIF cpos=UTF8GETPOS(inputstr,LEN(inputstr),0) THEN
          ' cursor is last char
          its=UTF8GETPOS(inputstr,LEN(inputstr),1-ibw)
          rgPRINT UTF8MID(inputstr,its,0,ibw-1)
          rgPRINT ESC("A")+UTF8MID(inputstr,cpos,0,1)+ESC("a")
        ELSE ' cursor is before last char
          IF cpos>1 THEN
            ' put out text in front of cursor
            rgPRINT UTF8MID(inputstr,its,0,UTF8GETDIFF(inputstr,cpos,its))
          ENDIF
          rgPRINT ESC("A")+UTF8MID(inputstr,cpos,0,1)+ESC("a")
          i=UTF8GETDIFF(inputstr,its,cpos)
          rgPRINT UTF8MID(inputstr,cpos,1,ibw-i-1)
        ENDIF
        ' put out scrolling arrows if necessary
        i=UTF8GETDIFF(inputstr,its,LEN(inputstr))
        IF i>=ibw AND cpos<=LEN(inputstr) THEN rgPRINT AT(ibx+ibw-1,iby)+CHR(26,36) ' show right arrow
        IF its>1 THEN rgPRINT AT(ibx, iby)+CHR(26,35) ' show left arrow
      ELSE ' string is not broader then input box
        IF cpos>LEN(inputstr) THEN
          ' put cursor after string
          rgPRINT AT(ibx, iby)+inputstr+CHR(27,65,32,27,97)
        ELSEIF cpos>1 THEN
          ' put cursor in the middle
          rgPRINT AT(ibx, iby)+UTF8LEFT(inputstr,cpos,-1)
          rgPRINT ESC("A")+UTF8MID(inputstr,cpos,0,1)+ESC("a")+UTF8MID(inputstr,cpos,1)
        ELSE
          ' put cursor at the head
          rgPRINT AT(ibx, iby)+ESC("A")+UTF8MID(inputstr,1,0,1)+ESC("a")+UTF8MID(inputstr,1,1)
        ENDIF
      ENDIF
    ENDIF
    SLEEP 10 ' for CPU idle
    rgSCREENSHOW
  LOOP

  IF retrogramousepointer<>0 THEN
    ' return to previous mouse pointer (if has been visible)
    rgSETMOUSE mouseform
  ENDIF

  rgPRINT AT(ibx, iby)+SPACE(ibw) ' clear "input field"
  IF key=CHR(13) THEN ' Enter
    ' remove cursor and arrows
    rgPRINT AT(ibx, iby)+UTF8MID(inputstr, its, 0, ibw)
  ELSE ' Escape / aborted
    inputstr=""
  ENDIF

  ' always perform a "return" after ("carriage return"+line feed) 
  rgPRINT AT(ibx, iby)+CHR(13)
END SUB
